/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2004-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Needed for snprintf */
#define _XOPEN_SOURCE	500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/asn1.h>
#include <openssl/asn1_mac.h>
#include <openssl/asn1t.h>
#include <openssl/x509v3.h>
#include <openssl/objects.h>
#include <openssl/err.h>

#include "_verify_proxy_certinfo.h"
#include "_verify_log.h"
#include "verify_x509_datatypes.h"


/* The code in this file should really be transformed along the lines of the
 * KTH's implementation in OpenSSL in crypto/x509v3/v3_pcia.c,
 * crypto/x509v3/v3_pci.c (and crypto/x509v3/x509v3.h).
 * For now we stick to the VOMS based code which is using the macros from
 * asn1_mac.h which cause unpreventable compiler warnings )-: */

/*****************************************************************************/
/* Defines                                                                   */
/*****************************************************************************/

/* error handling */
#define ASN1_F_D2I_PROXYPOLICY          451
#define ASN1_F_D2I_PROXYCERTINFO        431


/*****************************************************************************/
/* Internal prototypes                                                       */
/*****************************************************************************/

/* Define PROXYPOLICY ASN1 representation, make sure this is consistent with
 * OpenSSL in crypto/x509v3/v3_pcia.c for PROXY_POLICY */
ASN1_SEQUENCE(PROXYPOLICY) = {
    ASN1_SIMPLE(PROXYPOLICY, policy_language, ASN1_OBJECT),
    ASN1_OPT(PROXYPOLICY, policy, ASN1_OCTET_STRING)
} ASN1_SEQUENCE_END(PROXYPOLICY)
IMPLEMENT_ASN1_FUNCTIONS(PROXYPOLICY)

/* Define PROXYCERTINFO ASN1 representation, make sure this is consistent with
 * OpenSSL in crypto/x509v3/v3_pcia.c for PROXY_CERT_INFO_EXTENSION */
ASN1_SEQUENCE(PROXYCERTINFO) = {
    ASN1_OPT(PROXYCERTINFO, path_length, ASN1_INTEGER),
    ASN1_SIMPLE(PROXYCERTINFO, policy, PROXY_POLICY)
} ASN1_SEQUENCE_END(PROXYCERTINFO)
IMPLEMENT_ASN1_FUNCTIONS(PROXYCERTINFO)



/* DER (pp) to internal (a) */
static PROXYPOLICY *d2i_myPROXYPOLICY(PROXYPOLICY **a,
                                      const unsigned char **pp, long length);
/* internal (policy) to DER (pp) */
static int i2d_myPROXYPOLICY(PROXYPOLICY * policy, unsigned char **pp);


/* DER (pp) to internal (cert_info) for GT3: tries first GT3 proper, than RFC */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v3(PROXYCERTINFO **cert_info,
                                             const unsigned char **pp,
                                             long length);
/* DER (pp) to internal (cert_info) for Globus-compatible only GT3 */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v3_proper(PROXYCERTINFO **cert_info,
					            const unsigned char **pp,
					            long length);
/* DER (pp) to internal (cert_info) for RFC */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v4(PROXYCERTINFO **cert_info,
                                             const unsigned char **pp,
                                             long length);


/* internal (proxycertinfo) to DER (pp) for GT3 */
static int i2d_myPROXYCERTINFO_v3(PROXYCERTINFO *proxycertinfo,
                                  unsigned char ** pp);
/* internal (proxycertinfo) to DER (pp) for RFC */
static int i2d_myPROXYCERTINFO_v4(PROXYCERTINFO *proxycertinfo,
                                  unsigned char **pp);


/* internal (ext) to string (return value) */
static char *myproxycertinfo_i2s(const struct v3_ext_method *method,
				 void *ext);
/* string (data) to internal (ctx?) */
static void *myproxycertinfo_s2i(const struct v3_ext_method *method,
                                 struct v3_ext_ctx *ctx,
                                 const char *data);


/*****************************************************************************/
/* Function definitions                                                      */
/*****************************************************************************/

/* Adds a GT3 old-style proxy extension to the list of known extensions */
int init_GT3_proxy_extension(void)  {
    static X509V3_EXT_METHOD pcert;

    /* Proxy Certificate Extension's related objects */
    ERR_clear_error();
    OBJ_create(GLOBUS_PROXY_V3_OID, GLOBUS_PROXY_V3_SN, GLOBUS_PROXY_V3_LN);

    memset(&pcert, 0, sizeof(pcert));
    pcert.ext_nid = OBJ_txt2nid(GLOBUS_PROXY_V3_OID);
    pcert.ext_flags = 0;
    pcert.ext_new  = (X509V3_EXT_NEW) PROXYCERTINFO_new;
    pcert.ext_free = (X509V3_EXT_FREE)PROXYCERTINFO_free;
    pcert.d2i      = (X509V3_EXT_D2I) d2i_myPROXYCERTINFO_v3;  /* DER to internal */
    pcert.i2d      = (X509V3_EXT_I2D) i2d_myPROXYCERTINFO_v3;  /* internal to DER */
    pcert.i2s      = (X509V3_EXT_I2S) myproxycertinfo_i2s;  /* internal to string */
    pcert.s2i      = (X509V3_EXT_S2I) myproxycertinfo_s2i;  /* string to internal */
    pcert.v2i      = (X509V3_EXT_V2I) NULL; /* STACK_OF(CONF_VALUE) to internal */
    pcert.r2i      = (X509V3_EXT_R2I) NULL; /* internal to human readable */
    pcert.i2v      = (X509V3_EXT_I2V) NULL; /* internal to STACK_OF(CONF_VALUE) */
    pcert.i2r      = (X509V3_EXT_I2R) NULL; /* human readable to internal */

    X509V3_EXT_add(&pcert);

    return 0;
}

/* Adds an RFC proxy extension to the list of known extensions */
int init_RFC_proxy_extension(void)  {
    static X509V3_EXT_METHOD pcert;

    ERR_clear_error();
    OBJ_create(RFC_PROXY_OID, RFC_PROXY_SN, RFC_PROXY_LN);

    memset(&pcert, 0, sizeof(pcert));
    pcert.ext_nid = OBJ_txt2nid(RFC_PROXY_OID);
    pcert.ext_flags = 0;
    pcert.ext_new  = (X509V3_EXT_NEW) PROXYCERTINFO_new;
    pcert.ext_free = (X509V3_EXT_FREE)PROXYCERTINFO_free;
    pcert.d2i      = (X509V3_EXT_D2I) d2i_myPROXYCERTINFO_v4;  /* DER to internal */
    pcert.i2d      = (X509V3_EXT_I2D) i2d_myPROXYCERTINFO_v4;  /* internal to DER */
    pcert.i2s      = (X509V3_EXT_I2S) myproxycertinfo_i2s;  /* internal to string */
    pcert.s2i      = (X509V3_EXT_S2I) myproxycertinfo_s2i;  /* string to internal */
    pcert.v2i      = (X509V3_EXT_V2I) NULL; /* STACK_OF(CONF_VALUE) to internal */
    pcert.r2i      = (X509V3_EXT_R2I) NULL; /* internal to human readable */
    pcert.i2v      = (X509V3_EXT_I2V) NULL; /* internal to STACK_OF(CONF_VALUE) */
    pcert.i2r      = (X509V3_EXT_I2R) NULL; /* human readable to internal */

    X509V3_EXT_add(&pcert);

    return 0;
}

/******************************************************************************
 * Internal functions
 *****************************************************************************/

/* Convert DER representation of a PROXYPOLICY (pp) to an internal
 * representation (a) */
static PROXYPOLICY *d2i_myPROXYPOLICY(PROXYPOLICY **a,
				      const unsigned char **pp, long length)
{
    /* Note: M_ASN1_D2I_vars() defines an ASN1_const_CTX struct c. The macros
     * M_ASN1_D2I_get_opt() and M_ASN1_D2I_get_IMP_opt() call M_ASN1_next and
     * M_ASN1_next_prev, which try to assign values to * (const char *), i.e. to
     * const chars. They therefore cast to (unsigned char *) which rightfully
     * leads to compiler warnings (at least for gcc). See also comment from
     * Richard Levitte in asn1_mac.h above M_ASN1_next */
    M_ASN1_D2I_vars(a, PROXYPOLICY *, PROXYPOLICY_new);
    
    M_ASN1_D2I_Init();
    M_ASN1_D2I_start_sequence();
    M_ASN1_D2I_get(ret->policy_language, d2i_ASN1_OBJECT);

    /* need to try getting the policy using
     *     a) a call expecting no tags
     *     b) a call expecting tags
     * one of which should succeed
     */
    
    M_ASN1_D2I_get_opt(ret->policy,
                       d2i_ASN1_OCTET_STRING,
                       V_ASN1_OCTET_STRING);
    M_ASN1_D2I_get_IMP_opt(ret->policy,
                           d2i_ASN1_OCTET_STRING,
                           0,
                           V_ASN1_OCTET_STRING);
    M_ASN1_D2I_Finish(a, 
                      PROXYPOLICY_free, 
                      ASN1_F_D2I_PROXYPOLICY);
}

/* Convert an internal representation of a PROXYPOLICY (policy) to a DER
 * representation (pp) */
static int i2d_myPROXYPOLICY(PROXYPOLICY * policy, unsigned char **pp) {
    M_ASN1_I2D_vars(policy);

    M_ASN1_I2D_len(policy->policy_language, i2d_ASN1_OBJECT);

    if(policy->policy)
	M_ASN1_I2D_len(policy->policy, i2d_ASN1_OCTET_STRING);

    M_ASN1_I2D_seq_total();
    M_ASN1_I2D_put(policy->policy_language, i2d_ASN1_OBJECT);

    if(policy->policy)
	M_ASN1_I2D_put(policy->policy, i2d_ASN1_OCTET_STRING);

    M_ASN1_I2D_finish();
}


/* Convert DER representation of a GT3 PROXYCERTINFO (pp) to an internal
 * representation (cert_info) */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v3(PROXYCERTINFO **cert_info,
					     const unsigned char **pp,
					     long length)
{
    const char *logstr="d2i_myPROXYCERTINFO_v3";
    PROXYCERTINFO *rc=d2i_myPROXYCERTINFO_v3_proper(cert_info, pp, length);

    /* Note: Java VOMS library produces GT3 proxies with a RFC-style
     * PROXYCERTINFO structure. */
    if (rc) return rc;

    /* 'official' GT3 format failed, try RFC format */
    verify_log(L_WARN, "%s: trying to use proper GT3 proxycertinfo failed, "
	       "using RFC instead", logstr);
    return d2i_myPROXYCERTINFO_v4(cert_info, pp, length);
}

/* Convert DER representation of a proper GT3 PROXYCERTINFO (pp) to an internal
 * representation (cert_info) */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v3_proper(PROXYCERTINFO **cert_info,
					            const unsigned char **pp,
					            long length)
{
    /* See remark in d2i_myPROXYPOLICY() about M_ASN1_D2I_vars() and
     * compiler warnings */
    M_ASN1_D2I_vars(cert_info, PROXYCERTINFO *, PROXYCERTINFO_new);

    M_ASN1_D2I_Init();
    M_ASN1_D2I_start_sequence();

    M_ASN1_D2I_get((ret->policy), d2i_myPROXYPOLICY);

    M_ASN1_D2I_get_EXP_opt(ret->path_length, d2i_ASN1_INTEGER, 1);

    M_ASN1_D2I_Finish(cert_info, PROXYCERTINFO_free, ASN1_F_D2I_PROXYCERTINFO);
}

/* Convert DER representation of a RFC PROXYCERTINFO (pp) to an internal
 * representation (cert_info) */
static PROXYCERTINFO *d2i_myPROXYCERTINFO_v4(PROXYCERTINFO **cert_info,
					     const unsigned char **pp,
					     long length)
{
    /* See remark in d2i_myPROXYPOLICY() about M_ASN1_D2I_vars() and compiler
     * warnings */
    M_ASN1_D2I_vars(cert_info, PROXYCERTINFO *, PROXYCERTINFO_new);

    M_ASN1_D2I_Init();
    M_ASN1_D2I_start_sequence();

    M_ASN1_D2I_get_EXP_opt(ret->path_length, d2i_ASN1_INTEGER, 1);
    
    M_ASN1_D2I_get_opt(ret->path_length, d2i_ASN1_INTEGER, V_ASN1_INTEGER);

    M_ASN1_D2I_get((ret->policy),d2i_myPROXYPOLICY);

    M_ASN1_D2I_Finish(cert_info, PROXYCERTINFO_free, ASN1_F_D2I_PROXYCERTINFO);
}

/* Convert an internal representation of a GT3 PROXYCERTINFO (proxycertinfo) to
 * a DER representation (pp) */
static int i2d_myPROXYCERTINFO_v3(PROXYCERTINFO *proxycertinfo,
				  unsigned char ** pp) 
{
    int v1;

    M_ASN1_I2D_vars(proxycertinfo);
    
    v1 = 0;

    M_ASN1_I2D_len(proxycertinfo->policy, i2d_myPROXYPOLICY);

    M_ASN1_I2D_len_EXP_opt(proxycertinfo->path_length,i2d_ASN1_INTEGER, 1, v1);
    M_ASN1_I2D_seq_total();
    M_ASN1_I2D_put(proxycertinfo->policy, i2d_myPROXYPOLICY);
    M_ASN1_I2D_put_EXP_opt(proxycertinfo->path_length, i2d_ASN1_INTEGER, 1, v1);
    M_ASN1_I2D_finish();
}

/* Convert an internal representation of a RFC PROXYCERTINFO (proxycertinfo) to
 * a DER representation (pp) */
static int i2d_myPROXYCERTINFO_v4(PROXYCERTINFO *proxycertinfo,
				  unsigned char **pp) 
{
    M_ASN1_I2D_vars(proxycertinfo);

    if(proxycertinfo->path_length)
        M_ASN1_I2D_len(proxycertinfo->path_length, i2d_ASN1_INTEGER);
    
    M_ASN1_I2D_len(proxycertinfo->policy, i2d_myPROXYPOLICY);

    M_ASN1_I2D_seq_total();
    if(proxycertinfo->path_length)
        M_ASN1_I2D_put(proxycertinfo->path_length, i2d_ASN1_INTEGER);
    M_ASN1_I2D_put(proxycertinfo->policy, i2d_myPROXYPOLICY);
    M_ASN1_I2D_finish();
}

/* Convert an internal representation of a PROXYCERTINFO (ext) to a string
 * representation (return value). Probably not used in this verify library */
static char *myproxycertinfo_i2s(const struct v3_ext_method *method,
				 void *ext) {
    PROXYCERTINFO *pci = NULL;
    char *encoding = NULL;
    char *output = NULL;
    PROXYPOLICY *pp;
    int dooid = 0;
    char oid[256];
    long j;
    char *buffer=NULL;
    int len, len2;

    pci = (PROXYCERTINFO *)ext;
 
    if (!pci)
	return strdup("");

    if (pci->path_length) {
        j = ASN1_INTEGER_get(pci->path_length);

	len  = snprintf(output, 0, "%lX", j);
	len2 = snprintf(output, 0, "Path Length Constraint: %s%lX\n",
			( (len%2==1) ? "0" : ""), j);
	if ( len2<0 || (output=malloc((size_t)(len2+1))) == NULL )
	    return NULL;
	
        snprintf(output, (size_t)(len2+1), "Path Length Constraint: %s%lX\n",
		 ( (len%2==1) ? "0" : ""), j);
        free(buffer);
    }
    else
        output = strdup("Path Length Constraint: unlimited\n");

    pp = pci->policy;

    if (pp && i2t_ASN1_OBJECT(oid, 256, pp->policy_language))
        dooid  = 1;

    len = snprintf(encoding, 0, "%sPolicy Language: %s%s%s%s\n", 
		   output, 
		   ( dooid ? oid : ""), 
		   ( (pp && pp->policy) ? "\nPolicy Text: " : ""), 
		   ( (pp && pp->policy) ?
		        (char*)ASN1_STRING_data(pp->policy) :
		        ""),
		   ( (pp && pp->policy) ? "\n" : ""));

    if ( len<0 || (encoding=malloc((size_t)(len+1))) == NULL )   {
	free(output);
	return NULL;
    }

    snprintf(encoding, (size_t)(len+1), "%sPolicy Language: %s%s%s%s\n", 
	     output, 
	     ( dooid ? oid : ""), 
	     ( (pp && pp->policy) ? "\nPolicy Text: " : ""), 
	     ( (pp && pp->policy) ?
	          (char*)ASN1_STRING_data(pp->policy) :
	          ""),
	     ( (pp && pp->policy) ? "\n" : ""));
    free(output);

    return encoding;
}

/* Convert an string representation of a PROXYCERTINFO (data) to an internal
 * PROXYCERTINFO representation (return value). Probably not used in this verify
 * library */
static void *myproxycertinfo_s2i(const struct v3_ext_method *method,
				 struct v3_ext_ctx *ctx,
				 const char *data)
{
    /* Should parse input string and create a PROXYCERTINFO struct */
    PROXYCERTINFO *pci=PROXYCERTINFO_new();
    verify_error("myproxycertinfo_s2i", "This function is not implemented");

    return pci;
}
